#include <stdio.h>
#include <string.h>

#include "config.h"
#include "imapfilter.h"
#include "buffer.h"


extern conn_t connpri, connaux;

buffer_t obuf;			/* Output buffer. */

static unsigned int tag = 0x1000;	/* Every IMAP command is prefixed with
					 * a unique [:alnum:] string. */


unsigned int send_command(conn_t * conn, char *cmd);
void send_command_cont(conn_t * conn, char *cmd);


/*
 * Send to server data; a command.
 */
unsigned int
send_command(conn_t * conn, char *cmd)
{
#ifdef DEBUG
	fprintf(stderr, "debug: sending command (%s):\n\n%s\n",
	    (conn == &connpri ? "primary" : "auxiliary"), cmd);
#endif
	verbose("%s: %s", (conn == &connpri ? "C" : "c"), cmd);

	socket_write(conn, cmd);

	if (tag == 0xFFFF)
		tag = 0x0FFF;

	return tag++;
}


#ifdef CRAM_MD5
/*
 * Send to server data: a continuation command.
 */
void
send_command_cont(conn_t * conn, char *cmd)
{
#ifdef DEBUG
	fprintf(stderr, "debug: sending command (%s):\n\n%s\r\n\n",
	    (conn == &connpri ? "primary" : "auxiliary"), cmd);
#endif

	socket_write(conn, cmd);
	socket_write(conn, "\r\n");
}
#endif


#ifdef DEBUG
/*
 * IMAP NOOP: does nothing always succeeds.
 */
int
imap_noop(conn_t * conn)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("NOOP") + 12);

	snprintf(obuf.data, obuf.size, "%04X NOOP\r\n", tag);

	return send_command(conn, obuf.data);
}
#endif


/*
 * IMAP CAPABILITY: requests listing of capabilities that the server supports.
 */
int
imap_capability(conn_t * conn)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("CAPABILITY") + 12);

	snprintf(obuf.data, obuf.size, "%04X CAPABILITY\r\n", tag);

	return send_command(conn, obuf.data);
}


/*
 * IMAP NAMESPACE: discovers the prefix and delimeter of namespaces used by
 * the server for mailboxes (RFC 2342).
 */
int
imap_namespace(conn_t * conn)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("NAMESPACE") + 12);

	snprintf(obuf.data, obuf.size, "%04X NAMESPACE\r\n", tag);

	return send_command(conn, obuf.data);
}


/*
 * IMAP LOGOUT: informs server that client is done.
 */
int
imap_logout(conn_t * conn)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("LOGOUT") + 12);

	snprintf(obuf.data, obuf.size, "%04X LOGOUT\r\n", tag);

	return send_command(conn, obuf.data);
}


/*
 * IMAP STARTTLS: begin TLS negotiation.
 */
int
imap_starttls(conn_t * conn)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("STARTTLS") + 12);

	snprintf(obuf.data, obuf.size, "%04X STARTTLS\r\n", tag);

	return send_command(conn, obuf.data);
}


#ifdef CRAM_MD5
/*
 * IMAP AUTHENTICATE: indicates authentication mechanism and performs an
 * authentication protocol exchange.
 */
int
imap_authenticate(conn_t * conn, char *auth, int cont)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("AUTHENTICATE") + 12);

	if (!cont) {
		snprintf(obuf.data, obuf.size, "%04X AUTHENTICATE %s\r\n",
		    tag, auth);
		return send_command(conn, obuf.data);
	} else {
		send_command_cont(conn, auth);
		return 0;
	}
}
#endif


/*
 * IMAP LOGIN: identifies client to server.
 */
int
imap_login(conn_t * conn, char *user, char *pass)
{
	int r, n;
	char *sbuf;

	n = strlen("LOGIN") + strlen(user) + strlen(pass) + 18;
	sbuf = (char *)smalloc(n);

	snprintf(sbuf, n, "%04X LOGIN \"%s\" \"%s\"\r\n", tag, user, pass);

	r = send_command(conn, sbuf);

	sfree(sbuf);

	return r;
}


/*
 * IMAP LIST: returns a subset of names from the complete set of names
 * available to the client.
 *
int imap_list(conn_t *conn, char *refer, char *mbox)
{
	int r;
	
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("LIST") + strlen(refer) + strlen(mbox) + 18);
	
	snprintf(obuf.data, obuf.size, "%04X LIST \"%s\" \"%s\"\r\n", tag,
	    refer, mbox);

	r = send_command(conn, obuf.data);

	return r;
}*/


/*
 * IMAP SUBSCRIBE: adds the specified mailbox name to the server's
 * set of "active" or "subscribed" mailboxes.
 */
int
imap_subscribe(conn_t * conn, char *mbox)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("SUBSCRIBE") + strlen(mbox) + 15);

	snprintf(obuf.data, obuf.size, "%04X SUBSCRIBE \"%s\"\r\n", tag, mbox);

	return send_command(conn, obuf.data);
}


/*
 * IMAP EXAMINE: access a mailbox in READ-ONLY mode.
 *
int
imap_examine(conn_t *conn, char *mbox)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("EXAMINE") + strlen(mbox) + 15);

	snprintf(obuf.data, MEDIUM_CMD, "%04X EXAMINE \"%s\"\r\n", tag, mbox);

	return send_command(conn, obuf.data);
}*/


/*
 * IMAP SELECT: access a mailbox in READ-WRITE mode.
 */
int
imap_select(conn_t * conn, char *mbox)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("SELECT") + strlen(mbox) + 15);

	snprintf(obuf.data, obuf.size, "%04X SELECT \"%s\"\r\n", tag, mbox);

	return send_command(conn, obuf.data);
}


/*
 * IMAP STATUS: requests status of the indicated mailbox.
 */
int
imap_status(conn_t * conn, char *mbox, char *items)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("STATUS") + strlen(mbox) +
	    strlen(items) + 18);

	snprintf(obuf.data, obuf.size, "%04X STATUS \"%s\" (%s)\r\n", tag, mbox, items);

	return send_command(conn, obuf.data);
}


/*
 * IMAP CREATE: create mailbox.
 */
int
imap_create(conn_t * conn, char *mbox)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("CREATE") + strlen(mbox) + 14);

	snprintf(obuf.data, obuf.size, "%04X CREATE \"%s\"\r\n", tag, mbox);

	return send_command(conn, obuf.data);
}


/*
 * IMAP SEARCH: searches the mailbox for messages that match certain criteria.
 */
int
imap_search(conn_t * conn, char *charset, char *search)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("SEARCH CHARSET") + strlen(charset) +
	    strlen(search) + 15);

	if (*charset)
		snprintf(obuf.data, obuf.size, "%04X SEARCH CHARSET \"%s\" %s\r\n",
		    tag, charset, search);
	else
		snprintf(obuf.data, obuf.size, "%04X SEARCH %s\r\n", tag, search);

	return send_command(conn, obuf.data);
}


/*
 * IMAP FETCH: retrieves data associated with a message.
 */
int
imap_fetch(conn_t * conn, char *mesg, char *items)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("FETCH") + strlen(mesg) + strlen(items) + 14);

	snprintf(obuf.data, obuf.size, "%04X FETCH %s %s\r\n", tag, mesg, items);

	return send_command(conn, obuf.data);
}


/*
 * IMAP STORE: alters data associated with a message.
 */
int
imap_store(conn_t * conn, char *mesg, unsigned int mode, char *flags)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("STORE") + strlen(mesg) +
	    strlen("FLAGS.SILENT") + strlen(flags) + 18);

	snprintf(obuf.data, obuf.size, "%04X STORE %s %sFLAGS.SILENT (%s)\r\n",
	    tag, mesg, (mode == STORE_FLAG_REPLACE ? "" :
		mode == STORE_FLAG_ADD ? "+" : "-"), flags);

	return send_command(conn, obuf.data);
}


/*
 * IMAP COPY: copy messages to mailbox.
 */
int
imap_copy(conn_t * conn, char *mesg, char *mbox)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("COPY") + strlen(mesg) + strlen(mbox) + 16);

	snprintf(obuf.data, obuf.size, "%04X COPY %s \"%s\"\r\n", tag, mesg, mbox);

	return send_command(conn, obuf.data);
}


/*
 * IMAP APPEND: append message to the end of a mailbox.
 */
int
imap_append(conn_t * conn, char *mbox, unsigned int size)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("APPEND") + strlen(mbox) +
	    strlen(ultostr(size, 10)) + 18);

	snprintf(obuf.data, obuf.size, "%04X APPEND \"%s\" {%d}\r\n", tag, mbox,
	    size);

	return send_command(conn, obuf.data);
}



/*
 * IMAP CLOSE: delete messages and return to authenticated state.
 */
int
imap_close(conn_t * conn)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("CLOSE") + 12);

	snprintf(obuf.data, obuf.size, "%04X CLOSE\r\n", tag);

	return send_command(conn, obuf.data);
}


/*
 * IMAP EXPUNGE: permanently removes any messages with the \Deleted flag set.
 */
int
imap_expunge(conn_t * conn)
{
	reset_buffer(&obuf);
	check_buffer(&obuf, strlen("EXPUNGE") + 12);

	snprintf(obuf.data, obuf.size, "%04X EXPUNGE\r\n", tag);

	return send_command(conn, obuf.data);
}
